from .core.cobalt import Cobalt, Instance, StatusParent, Tunnel, download, get_tunnel


__all__ = [
    "Cobalt",
    "Instance",
    "Downloader",
    "Tunnel",
    "StatusParent",
    "download",
    "get_tunnel",
]



from aiohttp import ClientSession
from dotenv import load_dotenv
from os import getenv
from typing import (
    List,
    Dict,
    Union,
    Unpack,
    TypedDict,
    Literal,
    LiteralString,
    Any,
    Coroutine,
    Callable,
    AsyncGenerator,
)
from .misc import Translator, lprint, check_updates, cfg_value, StatusParent
from .client import RequestClient, _DownloadOptions
from .constants import (
    FALLBACK_INSTANCE,
    FALLBACK_INSTANCE_API_KEY,
    DEFAULT_UA,
    DEFAULT_TIMEOUT,
)
from . import exceptions
from time import time
from pathlib import Path
from .remux import remux
from sys import platform
from subprocess import run as srun
import re


class _CobaltParameters(TypedDict, total=False):
    instance: str
    api_key: str
    user_agent: str
    timeout: int
    translator: Translator = Translator()
    proxy: str
    session: ClientSession
    headers: Dict[str, str]
    debug: bool
    updates: bool


class _CobaltBodyOptions(TypedDict, total=False):
    url: str
    videoQuality: Literal[
        "max", "144", "240", "360", "480", "720", "1080", "1440", "2160", "4320"
    ]
    audioFormat: Literal["best", "mp3", "ogg", "wav", "opus"]
    audioBitrate: Literal["320", "256", "128", "96", "64", "8"]
    filenameStyle: Literal["classic", "pretty", "basic", "nerdy"]
    downloadMode: Literal["auto", "audio", "mute"]
    youtubeVideoCodec: Literal["h264", "av1", "vp9"]
    youtubeDubLang: LiteralString
    alwaysProxy: bool
    disableMetadata: bool
    tiktokFullAudio: bool
    tiktokH265: bool
    twitterGif: bool
    youtubeHLS: bool


class _CobaltDownloadOptions(TypedDict, total=False):
    url: str
    videoQuality: Literal[
        "max", "144", "240", "360", "480", "720", "1080", "1440", "2160", "4320"
    ] = "max"
    audioFormat: Literal["best", "mp3", "ogg", "wav", "opus"] = "best"
    audioBitrate: Literal["320", "256", "128", "96", "64", "8"] = "best"
    filenameStyle: Literal["classic", "pretty", "basic", "nerdy"] = "classic"
    downloadMode: Literal["auto", "audio", "mute"] = "auto"
    youtubeVideoCodec: Literal["h264", "av1", "vp9"] = "h264"
    youtubeDubLang: LiteralString = ""
    alwaysProxy: bool = False
    disableMetadata: bool = False
    tiktokFullAudio: bool = False
    tiktokH265: bool = False
    twitterGif: bool = False
    youtubeHLS: bool = False
    folder_path: str
    filename: str
    status_callback: Callable | Coroutine
    done_callback: Callable | Coroutine
    status_parent: str
    headers: Dict[str, str]
    timeout: int
    remux: bool
    show: bool
    open: bool


class Tunnel:
    url: str
    instance: "Instance" = None
    tunnel_id: str = None
    exp: int = None
    sig: str = None
    iv: str = None
    sec: str = None
    filename: str = None
    extension: str = None

    def __init__(
        self,
        data: dict,
        instance: "Instance" = None,
        auto_download: Unpack[_CobaltDownloadOptions] = None,
    ):
        url = data.get("url", None)
        if not url:
            raise exceptions.NoUrlInTunnelResponse(
                "No url found in response data while creating tunnel instance"
            )
        self.url = url
        self.instance = instance
        self.tunnel_id = (
            re.search(r"id=([^&]+)", url).group(1) if "id=" in url else None
        )
        self.exp = (
            re.search(r"exp=([^&]+)", url).group(1)[:-3] if "exp=" in url else None
        )
        self.sig = re.search(r"sig=([^&]+)", url).group(1) if "sig=" in url else None
        self.iv = re.search(r"iv=([^&]+)", url).group(1) if "iv=" in url else None
        self.sec = re.search(r"sec=([^&]+)", url).group(1) if "sec=" in url else None
        self.filename = data.get("filename", None)
        self.extension = self.filename.split(".")[-1]

    async def download(
        self, _remux: bool = False, **body: Unpack[_DownloadOptions]
    ) -> Path:
        if "url" not in body:
            body["url"] = self.url
        if "filename" not in body:
            body["filename"] = self.filename
        file_path = await self.instance.parent.request_client.download_from_url(**body)
        if _remux:
            file_path = remux(file_path)
        if body.get("open", False):
            if platform == "win32":
                from os import startfile

                startfile(file_path)
            elif platform == "darwin":
                srun(["open", file_path])
            else:
                srun(["xdg-open", file_path])
        if body.get("show", False):
            if platform == "win32":
                srun(
                    [
                        "explorer",
                        "/select,",
                        file_path,
                    ]
                )
            elif platform == "darwin":
                srun(["open", "-R", file_path])
            else:
                srun(
                    [
                        "xdg-open",
                        file_path,
                    ]
                )
        return file_path

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({(str(self.tunnel_id) + ', ' if self.tunnel_id else '') + self.url[:24] + '...'}"
            + (
                f", expires in {int(self.exp) - int(time())} seconds)"
                if self.exp and self.exp.isdigit()
                else ")"
            )
        )


class Instance:
    version: str = None
    url: str = None
    start_time: int = None
    duration_limit: int = None
    services: List[str] = None
    git: Dict[str, str] = None
    dump: Dict = None
    parent: "Cobalt" = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self.dump = kwargs
        if not self.url and kwargs.get("api", None):
            self.url = kwargs.get("api")
        if "http" not in self.url:
            self.url = f"https://{self.url}"
        if not self.parent:
            self.parent = Cobalt()

    async def get_instance_info(self, url: str = None):
        data = await self.parent.get(url or self.url)
        if not isinstance(data, dict):
            raise exceptions.FetchError(f"{self.url}: Failed to get instance data")
        _cobalt = data.get("cobalt", None)
        if isinstance(_cobalt, dict):
            self.version = _cobalt.get("version", None)
            self.url = _cobalt.get("url", None)
            self.start_time = _cobalt.get("start_time", None)
            self.duration_limit = _cobalt.get("duration_limit", None)
            self.services = _cobalt.get("services", None)
        return self

    async def get_tunnel(
        self, **body: Unpack[_CobaltBodyOptions]
    ) -> AsyncGenerator[Tunnel, None]:
        if not self.version == "unknown":
            try:
                await self.get_instance_info()
            except Exception:
                ...
        if len(re.findall("[&?]list=([^&]+)", body.get("url", ""))) > 0:
            from pytube import Playlist

            playlist = Playlist(body.get("url"))
            for i, item_url in enumerate(playlist.video_urls):
                if "music." in item_url:
                    item_url = item_url.replace("www", "music")
                body["url"] = item_url
                async for _tunnel in self.get_tunnel(**body):
                    yield _tunnel
        response = await self.parent.post(self.url, data=body)  # Pray for success!
        if not isinstance(response, dict):  # If your prayers not fulfilled
            if "<title>Just a moment...</title>" in response:
                raise exceptions.FailedToGetTunnel(
                    f"{self.url}: Cloudflare is blocking requests"
                )
            elif ">Sorry, you have been blocked</h1>" in response:
                raise exceptions.FailedToGetTunnel(
                    f"{self.url}: Site owner set that cloudflare is blocking your requests"
                )
            elif "ry again" in response:
                raise exceptions.FailedToGetTunnel(
                    f"{self.url}: Cloudflare failed to tunnel connection, try again later"
                )
            raise exceptions.FailedToGetTunnel(
                f"{self.url}: Reponse is not a dict: {response}"
            )
        if response.get("status", None) in ["error"]:
            raise exceptions.FailedToGetTunnel(
                f'{self.url}: {response.get("error", dict()).get("code", None)}'
            )
        if "url" not in response:
            raise exceptions.NoUrlInTunnelResponse(
                f"{self.url}: No url found in tunnel response: {response}"
            )
        yield Tunnel(response, instance=self)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.url}, {self.version if self.version else 'unknown'}, {len(self.services) if self.services else 0} services)"

    def __aiter__(self):
        return self


class Cobalt:
    instance: Union[Instance, str] = None
    fallback_instance: Union[Instance, str]
    api_key: str = None
    user_agent: str = None
    timeout: int = None
    translator: Translator = None
    proxy: str = None
    session: ClientSession = None
    headers: Dict[str, str] = None
    request_client: RequestClient = None
    solve_turnstile = True

    def __init__(self, **params: Unpack[_CobaltParameters]):
        self.__dict__.update(params)
        self.instances = []
        self.fallback_instance = Instance(
            url=FALLBACK_INSTANCE, api_key=FALLBACK_INSTANCE_API_KEY, parent=self
        )
        self.proxy = params.get("proxy", getenv("COBALT_PROXY", None))
        self.timeout = params.get("timeout", getenv("COBALT_TIMEOUT", DEFAULT_TIMEOUT))
        self.user_agent = params.get(
            "user_agent", getenv("COBALT_USER_AGENT", DEFAULT_UA)
        )
        self.api_key = params.get(
            "api_key", getenv("COBALT_API_KEY", FALLBACK_INSTANCE_API_KEY)
        )
        self.headers = params.get(
            "headers",
            {
                "User-Agent": self.user_agent,
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
        )
        self.request_client = RequestClient(
            api_key=self.api_key,
            session=self.session,
            headers=self.headers,
            timeout=self.timeout,
            proxy=self.proxy,
            user_agent=self.user_agent,
        )
        self.get = self.request_client.get
        self.post = self.request_client.post
        self.debug = (
            (
                lambda *args, **kwargs: lprint(
                    *[
                        ":gray:？ " + str(arg)
                        if not isinstance(arg, Exception)
                        else ":red::warning:  " + str(arg)
                        for arg in args
                    ],
                    **kwargs,
                )
            )
            if params.get("debug", getenv("COBALT_DEBUG", False))
            else lambda *args, **kwargs: ...
        )
        self.local_instance = Instance(
            url=getenv("COBALT_LOCAL_INSTANCE", "http://127.0.0.1:9000"),
            api_key=getenv("COBALT_LOCAL_INSTANCE_API_KEY", None),
            parent=self,
        )
        self.remux = remux
        if (
            "update" in params
            or int(cfg_value().get("update_check", 0)) + 60 * 60 * 3 < time()
        ):
            self.debug("Checking for updates...", end="\r")
            check_updates()
            cfg_value("update_check", str(int(time())))

    async def fetch_instances(self) -> List[Instance]:
        try:
            instances = await self.get(
                "https://instances.cobalt.best/api/instances.json"
            )
            if not isinstance(instances, list):
                raise exceptions.FetchError("Failed to fetch instances")
            self.instances = [
                Instance(parent=self, **instance) for instance in instances
            ] + [self.fallback_instance]
            try:
                self.debug("Fetching local instance info...", end="\r")
                await self.local_instance.get_instance_info()
                self.instances = [self.local_instance] + self.instances
            except Exception:
                self.debug("Didn't find local instance, skipping")
            self.debug(f"Fetched {len(self.instances)} instances")
            return self.instances
        except Exception as exc:
            self.debug(f"Failed to fetch instances: {exc}")
            self.instances = [self.fallback_instance]
            return self.instances

    async def get_tunnel(self, url: str, **body: Unpack[_CobaltBodyOptions]) -> Tunnel:
        tunnels = []
        for instance in await self.fetch_instances():
            try:
                async for tunnel in instance.get_tunnel(
                    url=url,
                    **{
                        key: value
                        for key, value in body.items()
                        if key in _CobaltBodyOptions.__annotations__.keys()
                    },
                ):
                    tunnels += [tunnel]
                return tunnels if len(tunnels) > 1 else tunnels[0]
            except Exception as exc:
                self.debug(exc)
        raise exceptions.AllInstancesFailed(
            f"Failed to get tunnel of {url} using any of {len(self.instances)} instances. If this issue persists, you can host your own local instance, more on it here: https://github.com/imputnet/cobalt/blob/main/docs/run-an-instance.md"
        )

    async def download(
        self, url: Union[str, Tunnel], **body: Unpack[_CobaltDownloadOptions]
    ) -> Path | List[Path]:
        if isinstance(url, Tunnel):
            return await self.request_client.download_from_url(url=url, **body)
        for instance in await self.fetch_instances():
            try:
                results = []
                self.debug(
                    f"Trying to download {url} using instance: {instance.url}", end="\r"
                )
                async for tunnel in instance.get_tunnel(
                    url=url,
                    **{
                        key: value
                        for key, value in body.items()
                        if key in _CobaltBodyOptions.__annotations__.keys()
                    },
                ):
                    self.debug(
                        f"Tunnel created, instance: {instance.url}, tunnel url: {tunnel.url}"
                    )
                    results.append(
                        await tunnel.download(
                            _remux=body.get("remux", False),
                            **{
                                key: value
                                for key, value in body.items()
                                if key in _CobaltDownloadOptions.__annotations__.keys()
                            },
                        )
                    )
                return results if len(results) > 1 else results[0]
            except Exception as exc:
                self.debug(exc)
        raise exceptions.AllInstancesFailed(
            f"Failed to download {url} using any of {len(self.instances)} instances. If this issue persists, you can host your own local instance, more on it here: https://github.com/imputnet/cobalt/blob/main/docs/run-an-instance.md"
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self.request_client.session and not self.request_client.session.closed:
            await self.request_client.session.close()

    def __setattr__(self, name: str, value: Any):
        if self.request_client and name in self.request_client.__dict__:
            self.request_client.__dict__[name] = value
        if isinstance(self.instance, Instance) and name in self.instance.__dict__:
            self.instance.__dict__[name] = value
        if isinstance(self.headers, dict):
            _ = ""
            for i, letter in enumerate(name):
                if _ == "" or name[i - 1] == "_":
                    _ += letter.upper()
                elif letter == "_":
                    _ += "-"
                else:
                    _ += letter
            if _ in self.headers:
                self.headers[_] = value
        self.__dict__[name] = value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"


class Downloader:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


DefaultCobalt = Cobalt()
download = DefaultCobalt.download
get_tunnel = DefaultCobalt.get_tunnel


load_dotenv()



from os import path, getenv, makedirs
from typing import TypedDict, Unpack
from shutil import get_terminal_size
from requests import get
from time import time
from rich.console import Console
import re


def get_cobalt_config_dir() -> str:
    _ = path.join(path.expanduser("~"), ".config", "cobalt")
    if not path.exists(_):
        makedirs(_, exist_ok=True)
    return _


cobalt_config_dir = get_cobalt_config_dir()


class Translator:
    language = getenv("LANG", "en")[:2]

    def __init__(self, language: str = None):
        self.language = language if language else self.language

    def translate(self, key: str, language: str = None) -> str:
        """Translate a key from the translation file."""
        language = language or self.language
        file = path.join(
            path.dirname(path.dirname(__file__)), "locales", f"{language}.txt"
        )
        if not path.exists(file):
            if language.upper() != "EN":
                return self.translate(key, "EN")
            return key
        with open(file) as f:
            for line in f:
                if "=" in line and line.split("=")[0].strip().upper() == key.upper():
                    return line[line.index("=") + 1 :].strip()
            if language.upper() != "EN":
                return self.translate(key, "EN")
            return key


def install_cobalt_container() -> None:
    """
    Installs and starts a local Cobalt instance using Docker.

    This function checks if Docker is installed on the system and installs it if not.
    It handles different installation procedures for Windows, and Linux.
    After ensuring Docker is installed, it downloads a docker-compose file and
    starts the Cobalt instance.

    The function prompts the user for confirmation before proceeding with the
    installation process.

    Note: The user may need to manually install Docker on unsupported operating
    systems or distributions.

    Raises:
        subprocess.CalledProcessError: If a command execution fails.
    """

    import platform
    import subprocess

    def is_docker_installed():
        try:
            subprocess.run(["docker", "--version"], check=True, capture_output=True)
            return True
        except (FileNotFoundError, subprocess.CalledProcessError):
            return False

    def install_docker_windows():
        print("Installing Docker Desktop on Windows...")
        subprocess.run(
            [
                "start",
                "",
                "/wait",
                "https://desktop.docker.com/win/main/amd64/Docker%20Desktop%20Installer.exe",
            ],
            shell=True,
            check=True,
        )
        print("Docker Desktop Installation complete")
        print("You might need to configure WSL2 if is not set")
        print("Please complete the setup process. Restart may be required.")

    def install_docker_macos():
        print("Installing Docker Desktop on macOS...")
        subprocess.run(
            ["open", "-W", "https://desktop.docker.com/mac/main/amd64/Docker.dmg"],
            check=True,
        )
        print("Docker Desktop Installation complete")
        print("Please complete the setup process. Restart may be required.")

    def install_docker_linux():
        print("Installing Docker Engine on Linux...")
        try:
            distro = platform.freedesktop_os_release().get("ID", None)
            if distro in ("ubuntu", "debian"):
                subprocess.run(["sudo", "apt", "update"], check=True)
                subprocess.run(
                    ["sudo", "apt", "install", "-y", "docker.io"], check=True
                )
                subprocess.run(
                    ["sudo", "systemctl", "enable", "docker", "--now"], check=True
                )
                print("Docker Engine Installation complete")
            elif distro in ("centos", "fedora", "rhel"):
                subprocess.run(["sudo", "yum", "update", "-y"], check=True)
                subprocess.run(["sudo", "yum", "install", "-y", "docker"], check=True)
                subprocess.run(
                    ["sudo", "systemctl", "enable", "docker", "--now"], check=True
                )
                print("Docker Engine Installation complete")
            elif distro in ("arch"):
                subprocess.run(["sudo", "pacman", "-Syu", "--noconfirm"], check=True)
                subprocess.run(
                    ["sudo", "pacman", "-S", "--noconfirm", "docker"], check=True
                )
                subprocess.run(
                    ["sudo", "systemctl", "enable", "docker", "--now"], check=True
                )
                print("Docker Engine Installation complete")
            else:
                print("Unsupported Linux distribution. Please install docker by hand.")
                return
        except subprocess.CalledProcessError:
            print("Error during installation on Linux. Please install docker by hand.")
            return

    while True:
        inp = str(
            input("You sure you want to install local cobalt instance? (y/n): ")
        ).lower()
        if inp == "y":
            break
        elif inp == "n":
            return

    if not is_docker_installed():
        inp = str(
            input("Docker is not installed. Do you want me to install it? (y/n): ")
        ).lower()
        if inp == "y":
            os_name = platform.system()
            if os_name == "Windows":
                install_docker_windows()
            elif os_name == "Darwin":
                install_docker_macos()
            elif os_name == "Linux":
                install_docker_linux()
            else:
                print(
                    "Unsupported operating system: " + os_name,
                    " Please install docker by hand.",
                )
            print(
                "Docker installation complete, you might need to restart your computer."
            )
        else:
            print("Please install docker by hand.")
            return

    with open(path.join(cobalt_config_dir, "docker-compose.yml"), "w+") as compose:
        compose.write(
            get(
                "https://raw.githubusercontent.com/imputnet/cobalt/refs/heads/main/docs/examples/docker-compose.example.yml"
            ).text
        )

    subprocess.run(["docker", "compose", "up", "-d"], check=True, cwd=cobalt_config_dir)
    print("Cobalt instance has been installed and started.")
    print(
        "If setup was successful, you can find the local cobalt instance at http://localhost:9000"
    )


class Terminal:
    console = Console()

    replaces = {
        ":accent:": "\033[96m",
        ":reset:": "\033[0m",
        ":end:": "\033[0m",
        ":bold:": "\033[1m",
        ":underline:": "\033[4m",
        ":italic:": "\033[3m",
        ":strikethrough:": "\033[9m",
        ":red:": "\033[31m",
        ":green:": "\033[32m",
        ":yellow:": "\033[33m",
        ":blue:": "\033[34m",
        ":magenta:": "\033[35m",
        ":cyan:": "\033[36m",
        ":purple:": "\033[35m",
        ":orange:": "\033[33m",
        ":pink:": "\033[35m",
        ":light_gray:": "\033[37m",
        ":dark_gray:": "\033[90m",
        ":lime:": "\033[92m",
        ":white:": "\033[37m",
        ":gray:": "\033[90m",
        ":bg_black:": "\033[40m",
        ":bg_red:": "\033[41m",
        ":bg_green:": "\033[42m",
        ":bg_yellow:": "\033[43m",
        ":bg_blue:": "\033[44m",
        ":bg_magenta:": "\033[45m",
        ":bg_cyan:": "\033[46m",
        ":bg_white:": "\033[47m",
    }

    pattern = re.compile(
        r"[\x1b\x9b\x9f][\[\]()\\]*[0-?]*[ -/]*[@-~]"
        r"|[\U00010000-\U0010ffff]"
        r"|[\u200d]"
        r"|[\u2640-\u2642]"
        r"|[\u2600-\u2b55]"
        r"|[\u23cf]"
        r"|[\u23e9]"
        r"|[\u231a]"
        r"|[\ufe0f]"  # dingbats
        r"|[\u3030]"
        "+",
        flags=re.UNICODE,
    )

    @classmethod
    def get_size(cls) -> tuple[int, int]:
        return get_terminal_size()

    @classmethod
    def apply_style(cls, text: str) -> str:
        for key, value in cls.replaces.items():
            text = text.replace(key, value)
        return text

    @classmethod
    def true_len(cls, text: str) -> int:
        text = cls.apply_style(text)
        _ = None
        for i, char in enumerate(text):
            if char == ":" and not _:
                _ = i
            elif char == ":" and _:
                text = text.replace(text[_ + 1 : i], "")
            elif char == " " and _:
                _ = None
        return len(
            re.sub(
                r"[\u001B\u009B][\[\]()#;?]*((([a-zA-Z\d]*(;[-a-zA-Z\d\/#&.:=?%@~_]*)*)?\u0007)|((\d{1,4}(?:;\d{0,4})*)?[\dA-PR-TZcf-ntqry=><~]))",
                "",
                text,
            )
        )

    @classmethod
    def lprint(cls, *args: str, right: bool = False, **kwargs) -> None:
        args = [
            cls.apply_style(
                str(arg) if not isinstance(arg, Exception) else ":red:" + str(arg)
            )
            for arg in args
        ]
        terminal_width = cls.get_size()[0]
        num_args = len(args)

        if "highlight" not in kwargs:
            kwargs["highlight"] = False

        if num_args == 0:
            return

        if num_args == 3:
            _center = args[1].center(terminal_width).rstrip()
            print(
                " " * (terminal_width - cls.true_len(args[2]))
                + cls.apply_style(args[2])
                + cls.apply_style(":end:"),
                end="\r",
            )
            cls.console.print(
                " " * ((len(_center) - cls.true_len(_center)) // 2)
                + _center
                + cls.apply_style(":end:"),
                end="\r",
                highlight=kwargs["highlight"],
            )
            cls.console.print(args[0], end="\r", highlight=kwargs["highlight"])
        elif num_args == 2:
            print(
                " " * (terminal_width - cls.true_len(args[1]))
                + cls.apply_style(args[1])
                + cls.apply_style(":end:"),
                end="\r",
            )
            cls.console.print(
                args[0] + cls.apply_style(":end:"),
                end="\r",
                highlight=kwargs["highlight"],
            )
        else:
            if right:
                print(
                    " " * (terminal_width - cls.true_len(args[0]))
                    + cls.apply_style(args[0])
                    + cls.apply_style(":end:"),
                    end="\r",
                )
            else:
                cls.console.print(
                    args[0].ljust(terminal_width),
                    end="\r",
                    highlight=kwargs["highlight"],
                )
        cls.console.print(cls.apply_style(":end:"), **kwargs)


lprint = Terminal.lprint
translate = Translator().translate
tl = translate


class StatusParent:
    total_size: int
    downloaded_size: int
    start_at: int
    time_passed: float
    file_path: str
    filename: str
    download_speed: int
    completed: bool

    def __init__(self) -> None:
        self.total_size = 0
        self.downloaded_size = 0
        self.start_at = 0
        self.time_passed = 0
        self.file_path = None
        self.filename = None
        self.download_speed = 0
        self.completed = False

    def __repr__(self) -> str:
        values = ", ".join(f"{key}={value!r}" for key, value in self.__dict__.items())
        return f"{self.__class__.__name__}({values})"


class _DownloadCallbackData(TypedDict):
    filename: str
    downloaded_size: int
    start_at: int
    time_passed: int | float
    file_path: str
    download_speed: int
    total_size: int
    iteration: int
    eta: int


class DefaultCallbacks:
    @classmethod
    async def status_callback(cls, **data: Unpack[_DownloadCallbackData]) -> None:
        percent_downloaded = (
            int((data.get("downloaded_size") / data.get("total_size")) * 100)
            if data.get("total_size", -1) != -1
            else -1
        )
        if percent_downloaded != -1:
            bar_length = 24
            completed_length = int(round(bar_length * percent_downloaded / float(100)))
            bar_fill = "▇" * completed_length
            bar_empty = "-" * (bar_length - completed_length)
            spinner = ["⢎⡰", "⢎⡡", "⢎⡑", "⢎⠱", "⠎⡱", "⢊⡱", "⢌⡱", "⢆⡱"]
            current_spin = spinner[data.get("iteration") % len(spinner)]
            lprint(
                f"⭳  :gray:{data.get('filename')} ",
                f":gray:[:green:{bar_fill}:gray:{bar_empty}]:end: :lime:{data.get('downloaded_size') / (1024 * 1024):.2f}:gray:/{data.get('total_size') / (1024 * 1024):.2f}MB :magenta:{data.get('download_speed') / (1024 * 1024):.2f}MB/s :cyan:{data.get('time_passed'):.2f}s :white::bold:{current_spin}",
                end="\r",
                highlight=False,
            )
        else:
            spinner = ["⢎⡰", "⢎⡡", "⢎⡑", "⢎⠱", "⠎⡱", "⢊⡱", "⢌⡱", "⢆⡱"]
            current_spin = spinner[data.get("iteration") % len(spinner)]
            lprint(
                f"⭳  :gray:{data.get('filename')} ",
                f":lime:{data.get('downloaded_size') / (1024 * 1024):.2f}MB :magenta:{data.get('download_speed') / (1024 * 1024):.2f}MB/s :cyan:{data.get('time_passed'):.2f}s :white::bold:{current_spin}",
                end="\r",
            )

    @classmethod
    async def done_callback(cls, **data: Unpack[_DownloadCallbackData]) -> None:
        file_size = path.getsize(data["file_path"]) / (1024 * 1024)
        lprint(
            f":green:✔  :white:{data['file_path']}",
            f":green:{file_size:.2f}MB :cyan:{data.get('time_passed'):.2f}s",
        )


def check_updates() -> bool:
    """
    Checks for updates of pybalt by comparing the current version to the latest version from pypi.org.

    Returns:
        bool: True if the check was successful, False otherwise.
    """
    from pkg_resources import get_distribution, DistributionNotFound
    from requests import get, exceptions

    try:
        current_version = get_distribution("pybalt").version
        response = get("https://pypi.org/pypi/pybalt/json", timeout=10)
        response.raise_for_status()
        data = response.json()
        last_version = data["info"]["version"]
        if last_version != current_version:
            lprint(
                tl("UPDATE_AVALIABLE").format(
                    last_version=last_version, current_version=current_version
                )
            )
            return False
    except DistributionNotFound:
        lprint(tl("PACKAGE_NOT_FOUND"))
    except exceptions.RequestException as e:
        lprint(tl("UPDATE_CHECK_FAIL").format(error=e))
    return True


def cfg_value(key: str = None, value: str = None) -> dict | str | None:
    cfg_path = path.join(cobalt_config_dir, "config.cfg")
    if not path.exists(cfg_path):
        with open(cfg_path, "w") as f:
            f.write("")
    if key is None:
        cfg_items = {}
        with open(cfg_path, "r") as f:
            for line in f:
                if "=" in line:
                    cfg_items[line.split("=")[0].strip()] = line.split("=")[1].strip()
        return cfg_items
    elif value is None:
        with open(cfg_path, "r") as f:
            for line in f:
                if line.startswith(key):
                    return line.split("=")[1].strip()
    else:
        cfg_content = []
        found = False
        with open(cfg_path, "r") as f:
            for line in f:
                if line.startswith(key):
                    line = f"{key}={value}\n"
                    found = True
                cfg_content.append(line)
        if not found:
            cfg_content.append(f"{key}={value}\n")
        with open(cfg_path, "w") as f:
            f.writelines(cfg_content)



from asyncio import run
from .core.cobalt import (
    _CobaltDownloadOptions,
    _CobaltParameters,
    Cobalt,
    lprint,
    cfg_value,
)
from typing import _LiteralGenericAlias
from os.path import exists
from time import time
import argparse


parser = argparse.ArgumentParser()
parser.add_argument("positional", nargs="?", type=str)
for key, value in {
    **_CobaltDownloadOptions.__annotations__,
    **_CobaltParameters.__annotations__,
}.items():
    try:
        if value is bool:
            if not any(
                arg.startswith(f"-{key[0]}") for arg in parser._option_string_actions
            ):
                parser.add_argument(
                    f"-{key[0]}{''.join([x for i, x in enumerate(key) if i > 0 and x.isupper()])}",
                    f"--{key}",
                    action="store_true",
                )
            else:
                parser.add_argument(
                    f"-{key[:2].lower()}",
                    f"--{key}",
                    action="store_true",
                )
        else:
            if not any(
                arg.startswith(f"-{key[0]}") for arg in parser._option_string_actions
            ):
                parser.add_argument(
                    f"-{key[0]}{''.join([x for i, x in enumerate(key) if i > 0 and x.isupper()])}",
                    f"--{key}",
                    type=value if not isinstance(value, _LiteralGenericAlias) else str,
                    choices=None
                    if not isinstance(value, _LiteralGenericAlias)
                    else value.__args__,
                )
            else:
                parser.add_argument(
                    f"-{key[:2].lower()}",
                    f"--{key}",
                    type=value if not isinstance(value, _LiteralGenericAlias) else str,
                    choices=None
                    if not isinstance(value, _LiteralGenericAlias)
                    else value.__args__,
                )
    except argparse.ArgumentError:
        if value is bool:
            parser.add_argument(f"--{key}", action="store_true")
        else:
            parser.add_argument(f"--{key}", type=value)


async def _():
    args = parser.parse_args()
    if args.positional and not args.url:
        args.url = args.positional
    cobalt = Cobalt(
        **{
            key: value
            for key, value in args.__dict__.items()
            if key in _CobaltParameters.__annotations__.keys() and value is not None
        }
    )
    if (
        args.positional
        and args.remux
        and exists(args.positional)
        and not args.positional.endswith(".txt")
    ):
        cobalt.remux(args.positional, keep_original=True)
    elif args.remux and not args.url:
        lprint(":red::warning: No URL or FILE PATH provided for remuxing")
    elif args.url:
        if exists(args.url):
            _urls = open(args.url).readlines()
        else:
            _urls = [args.url]
        for url in _urls:
            try:
                args.url = url
                await cobalt.download(
                    **{
                        key: value
                        for key, value in args.__dict__.items()
                        if key in _CobaltDownloadOptions.__annotations__.keys()
                        and value is not None
                    }
                )
            except Exception as exc:
                lprint(":red::warning: " + str(exc))
        if int(cfg_value().get("last_thank", 0)) < int(time()) - 60 * 60 * 2:
            cfg_value("last_thank", int(time()))
            lprint(
                ":green:Thanks for using pybalt:end:, if you enjoyed it please leave a star on GitHub :star: :accent:https://github.com/nichind/pybalt"
            )
    elif args.updates:
        lprint(":green:pybalt is up to date")
    else:
        lprint(":red::warning: No URL provided")


def main():
    run(_())


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
from pybalt.core import constants


def readme():
    with open("README.md", "r") as f:
        return f.read()


setup(
    name="pybalt",
    version=constants.VERSION,
    author="nichind",
    author_email="nichinddev@gmail.com",
    description="tool/python-module for downloading videos from YouTube, Twitter (X), Instagram, Reddit, Soundcloud and more! Powered by the @imputnet cobalt instances.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/nichind/pybalt",
    packages=find_packages(),
    package_data={
        "pybalt": ["locales/*.txt"],
    },
    install_requires=[x.strip() for x in open("requirements.txt").readlines()],
    keywords=[
        "downloader",
        "cobalt",
        "cobalt-cli",
        "youtube",
        "twitter",
        "x",
        "instagram",
        "reddit",
        "twitch",
        "bilibili",
        "download",
        "youtube-downloader",
        "twitter-downloader",
        "x-downloader",
        "instagram-downloader",
        "reddit-downloader",
        "twitch-downloader",
        "bilibili-downloader",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    entry_points={
        "console_scripts": [
            "pybalt=pybalt.__main__:main",
            "cobalt=pybalt.__main__:main",
        ],
    },
)



class BaseException(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


class FailedToGetTunnel(BaseException): ...


class NoUrlInTunnelResponse(BaseException): ...


class InvalidURL(BaseException): ...


class FetchError(BaseException): ...


class PageNotFound(BaseException): ...


class AllInstancesFailed(BaseException): ...


class DownloadError(BaseException): ...



from aiohttp import ClientSession
from .misc import Translator, DefaultCallbacks, StatusParent
from typing import Literal, Dict, Callable, Coroutine, Unpack, TypedDict
from asyncio import sleep, iscoroutinefunction, wait_for, TimeoutError
from .constants import DEFAULT_TIMEOUT
from time import time
from os import path, getenv, makedirs
from . import exceptions
from aiofiles import open as aopen
from pathlib import Path


class Response:
    text = None
    json = None
    headers = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return self.text if not self.json else self.json

    def __str__(self):
        return str(self.text)


class _DownloadOptions(TypedDict, total=False):
    url: str
    folder_path: str
    filename: str
    status_callback: Callable | Coroutine
    done_callback: Callable | Coroutine
    status_parent: str
    headers: Dict[str, str]
    timeout: int
    callback_rate: float
    proxy: str
    max_speed: int


class RequestClient:
    api_key: str = None
    user_agent: str = None
    timeout: int = None
    translator: Translator = None
    proxy: str = None
    session: ClientSession = None
    headers: Dict[str, str] = None
    verify_proxy: bool = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    async def request(
        self, url: str, request_type: Literal["get", "post"] = "get", **kwargs
    ) -> Response:
        retries = kwargs.get("retries", 0)
        if retries > 5:
            raise ""
        if self.proxy or kwargs.get("proxy", False):
            import requests

            requests.packages.urllib3.disable_warnings()  # Disable SSL warnings on proxy

            url = (
                f"{kwargs.get('base_url', self.base_url)}{url}"
                if not url.startswith("http")
                else url
            )
            proxies = {
                "http": self.proxy or kwargs.get("proxy", False),
                "https": self.proxy or kwargs.get("proxy", False),
            }
            if request_type == "get":
                response = requests.get(
                    url,
                    params=kwargs.get("params"),
                    headers=kwargs.get("headers", self.headers),
                    proxies=proxies,
                    verify=self.verify_proxy or kwargs.get("verify", False),
                    timeout=kwargs.get("timeout", self.timeout or DEFAULT_TIMEOUT),
                )
            else:
                response = requests.post(
                    url,
                    json=kwargs.get("data"),
                    headers=kwargs.get("headers", self.headers),
                    proxies=proxies,
                    verify=self.verify_proxy or kwargs.get("verify", False),
                    timeout=kwargs.get("timeout", self.timeout or DEFAULT_TIMEOUT),
                )
            if response.status_code == 429:
                kwargs["retries"] = retries + 1
                return await self.request(url, request_type, **kwargs)
            elif response.status_code == 404:
                raise exceptions.PageNotFound(f"{url}: Page not found")
            elif "REMOTE_ADDR = " in response.text:
                kwargs["retries"] = retries + 1
                return await self.request(url, request_type, **kwargs)
            _response = Response(
                headers=response.headers,
            )
            try:
                _response.text = response.text
                _response.json = response.json()
            except Exception:
                ...
            return _response
        else:
            session = (
                ClientSession(
                    headers=kwargs.get("headers", self.headers),
                )
                if not self.session or self.session.closed
                else self.session
            )
            try:
                url = (
                    f"{kwargs.get('base_url', self.base_url)}{url}"
                    if not url.startswith("http")
                    else url
                )
                async with (
                    session.get(
                        url,
                        params=kwargs.get("params"),
                        timeout=kwargs.get("timeout", self.timeout or DEFAULT_TIMEOUT),
                    )
                    if request_type == "get"
                    else session.post(
                        url,
                        json=kwargs.get("data"),
                        timeout=kwargs.get("timeout", self.timeout or DEFAULT_TIMEOUT),
                    )
                ) as response:
                    if response.status == 429:
                        retry_after = response.headers.get("Retry-After", 1)
                        await sleep(float(retry_after))
                        kwargs["retries"] = retries + 1
                        return await self.request(url, request_type, **kwargs)
                    elif response.status == 404:
                        raise exceptions.PageNotFound(f"{url}: Page not found")
                    _response = Response(headers=response.headers)
                    try:
                        _response.text = await response.text()
                        _response.json = await response.json()
                    except Exception:
                        ...
                    return _response
                kwargs["retries"] = retries + 1
                await sleep(1)
                return await self.request(url, request_type, **kwargs)
            finally:
                if kwargs.get("close", True):
                    await session.close()

    async def get(self, url: str, **kwargs) -> Response:
        response = await self.request(url, "get", **kwargs)
        if kwargs.get("text", False) is True:
            return response.text
        return response.json if response.json else response.text

    async def post(self, url: str, **kwargs) -> Response:
        response = await self.request(url, "post", **kwargs)
        if kwargs.get("text", False) is True:
            return response.text
        return response.json if response.json else response.text

    async def download_from_url(
        self,
        **options: Unpack[_DownloadOptions],
    ) -> Path:
        start_at = int(time())
        downloaded_size = 0
        destination_folder = options.get(
            "folder_path", getenv("COBALT_DOWNLOAD_FOLDER", None)
        ) or path.join(path.expanduser("~"), "Downloads")
        if "status_callback" not in options.keys():
            options["status_callback"] = DefaultCallbacks.status_callback
        if "done_callback" not in options.keys():
            options["done_callback"] = DefaultCallbacks.done_callback
        if not path.exists(destination_folder):
            makedirs(destination_folder)
        session = (
            ClientSession(
                headers=options.get("headers", self.headers),
            )
            if not self.session or self.session.closed
            else self.session
        )
        try:
            async with session.get(
                options.get("url"),
                proxy=options.get("proxy", None),
            ) as resp:
                if resp.status.__str__()[0] in ["4", "5"]:
                    raise exceptions.DownloadError(
                        "Failed to download file, status code: " + resp.status.__str__()
                    )
                total_size = int(resp.headers.get("Content-Length", -1))
                filename = options.get(
                    "filename",
                    (
                        resp.headers.get(
                            "Content-Disposition",
                            f'filename="{options.get("url").split("/")[-1].split("?")[0]}"',
                        )
                    )
                    .split('filename="')[1]
                    .split('"')[0],
                )
                file_path = path.join(
                    destination_folder,
                    filename,
                )
                async with aopen(file_path, "wb") as f:
                    last_callback = 0
                    last_size = 0
                    download_speed = 0
                    iteration = 0
                    last_callback = time()
                    max_speed = options.get("max_speed")
                    while True:
                        if total_size == -1:
                            chunk = await resp.content.read(1024 * 1024)
                            if not chunk:
                                break
                            # TODO: Implement non-blocking way to read buffer when connection is dropped and total file size is unknown
                            # try:
                            #     chunk = await wait_for(
                            #         resp.content.read(1024 * 1024),
                            #         0.512
                            #     )
                            #     if not chunk:
                            #         break
                            # except TimeoutError:
                            #     pass
                        else:
                            chunk = resp.content.read_nowait()
                        await f.write(chunk)
                        downloaded_size += len(chunk)
                        if time() - last_callback >= options.get(
                            "callback_rate", 0.128
                        ):
                            download_speed = (downloaded_size - last_size) / (
                                time() - last_callback
                                if time() - last_callback > 0
                                else 0.01
                            )
                            eta = (
                                (total_size - downloaded_size) / download_speed
                                if download_speed != 0
                                else 0
                            )
                            last_size = downloaded_size
                            if options.get("status_callback", None) is not None:
                                if iscoroutinefunction(options.get("status_callback")):
                                    await (options.get("status_callback"))(
                                        downloaded_size=downloaded_size,
                                        start_at=start_at,
                                        time_passed=round(time() - start_at, 2),
                                        file_path=file_path,
                                        filename=filename,
                                        download_speed=download_speed,
                                        total_size=total_size,
                                        iteration=iteration,
                                        eta=round(eta),
                                    )
                                else:
                                    (options.get("status_callback"))(
                                        downloaded_size=downloaded_size,
                                        start_at=start_at,
                                        time_passed=round(time() - start_at, 2),
                                        file_path=file_path,
                                        filename=filename,
                                        download_speed=download_speed,
                                        total_size=total_size,
                                        iteration=iteration,
                                        eta=round(eta),
                                    )
                            if options.get("status_parent", None):
                                if isinstance(
                                    options.get("status_parent"), StatusParent
                                ):
                                    options.get(
                                        "status_parent"
                                    ).downloaded_size = downloaded_size
                                    options.get("status_parent").start_at = start_at
                                    options.get("status_parent").time_passed = round(
                                        time() - start_at, 2
                                    )
                                    options.get("status_parent").file_path = file_path
                                    options.get("status_parent").filename = filename
                                    options.get(
                                        "status_parent"
                                    ).download_speed = download_speed
                                    options.get("status_parent").total_size = total_size
                                elif isinstance(options.get("status_parent"), dict):
                                    options.get("status_parent").update(
                                        downloaded_size=downloaded_size,
                                        start_at=start_at,
                                        time_passed=round(time() - start_at, 2),
                                        file_path=file_path,
                                        filename=filename,
                                        download_speed=download_speed,
                                        total_size=total_size,
                                    )
                                else:
                                    raise TypeError(
                                        "status_parent must be dict or StatusParent"
                                    )
                            last_callback = time()
                            iteration += 1
                        if downloaded_size >= total_size and total_size != -1:
                            break
                        if max_speed and len(chunk) > max_speed:
                            await sleep(
                                (len(chunk) - max_speed) / (max_speed / 1024 / 1024)
                            )
            if downloaded_size <= 1024:
                raise exceptions.DownloadError(
                    "Download failed, no data received. That's a problem with cobalt instance, try enabling youtubeHLS, if this issue persist host your own local instance."
                )
            if options.get("status_parent", None):
                if isinstance(options.get("status_parent"), StatusParent):
                    options.get("status_parent").downloaded_size = downloaded_size
                    options.get("status_parent").start_at = start_at
                    options.get("status_parent").time_passed = round(
                        time() - start_at, 2
                    )
                    options.get("status_parent").file_path = file_path
                    options.get("status_parent").filename = filename
                    options.get("status_parent").download_speed = download_speed
                    options.get("status_parent").total_size = total_size
                    options.get("status_parent").completed = True
                elif isinstance(options.get("status_parent"), dict):
                    options.get("status_parent").update(
                        downloaded_size=downloaded_size,
                        start_at=start_at,
                        time_passed=round(time() - start_at, 2),
                        file_path=file_path,
                        filename=filename,
                        download_speed=download_speed,
                        total_size=total_size,
                        completed=True,
                    )
                else:
                    raise TypeError("status_parent must be dict or StatusParent")
            if options.get("done_callback", None):
                if iscoroutinefunction(options.get("done_callback")):
                    await (options.get("done_callback"))(
                        downloaded_size=downloaded_size,
                        start_at=start_at,
                        time_passed=round(time() - start_at, 2),
                        file_path=file_path,
                        filename=filename,
                        total_size=path.getsize(file_path),
                    )
                else:
                    (options.get("done_callback"))(
                        downloaded_size=downloaded_size,
                        start_at=start_at,
                        time_passed=round(time() - start_at, 2),
                        file_path=file_path,
                        filename=filename,
                        total_size=path.getsize(file_path),
                    )
        except Exception as exc:
            raise exc
        finally:
            if options.get("close", True):
                await session.close()
        return Path(file_path)

    def __setattr__(self, name, value):
        self.__dict__[name] = value
        if name in self.headers:
            self.headers[name] = value
